from box import Box
import yaml
from pathlib import Path
import os
import logging.config
from pathlib import Path
from pydelling.utils.configuration_utils import get_config_path


def read_config(config_file: Path="./local_config.yaml"):
    """
    Reads the configuration file
    :param config_file:
    :return:
    """
    with open(config_file) as file:
        context = yaml.load(file, Loader=yaml.FullLoader)
    return Box(context, default_box=True)


_config_file = list(Path.cwd().glob("*config*"))
if not _config_file:
    _config_file = [get_config_path() / 'local_config.yaml']
    logging.warning('Using default configuration file (not the user defined one)')
assert len(_config_file) >= 1, "Please provide a configuration file that has a '*config.yaml' name structure"
config = read_config(config_file=_config_file[0])

# Add global configuration
if not config.globals.is_globals_loaded:
    with open(get_config_path() / "global_config.yaml", "r") as yml_file:
        local_yaml_file = yaml.safe_load(yml_file)
        config.globals = local_yaml_file
        config.globals.is_globals_loaded = True

# Allow to override the global configuration with a local file
for key in config:
    if key in config.globals:
        if isinstance(config[key], dict):
            config.globals[key].update(config[key])
        else:
            config.globals[key] = config[key]

os.makedirs(config.path.logs if config.path.logs else Path().cwd() / "logs", exist_ok=True)
with open(Path(__file__).parent / "logger_config.yaml", "r") as ymlfile:
    log_config = yaml.safe_load(ymlfile)
logging.config.dictConfig(log_config)


# Fix logging issue caused by streamlit
import streamlit.logger
loggers = [handler for handler in logging.root.handlers if isinstance(handler, logging.StreamHandler)]
strange_logger = loggers[-1]
strange_logger.setLevel(logging.ERROR)






"""
Base interface for a reader class
"""
import numpy as np
import logging
from pathlib import Path
import matplotlib.pyplot as plt
import pandas as pd

logger = logging.getLogger(__name__)
import pydelling.utils.SubFishModule as subfish
import seaborn as sns

class BaseReader:
    data: np.ndarray  # Hint of self.data array
    info: dict

    def __init__(self, filename=None, header=False, **kwargs):
        self.filename = Path(filename)
        self.info = {"reader": {}}
        self.data = None
        self.header = header
        self.__dict__.update(kwargs)
        self.open_file(filename, **kwargs)

    def read_file(self, opened_file):
        """
        Reads the data and stores it inside the class
        :return:
        """
        pass

    def open_file(self, filename, **kwargs):
        with open(filename) as opened_file:
            if self.header:
                opened_file.readline()  # For now, skips the header if it has
            self.read_file(opened_file)
        self.build_info()

    def read_header(self, opened_file):
        """
        Reads the header of the file
        :return:
        """
        pass

    def get_data(self) -> np.ndarray:
        """
        Outputs the read data
        :return:
        """
        return np.array(0)

    def build_info(self):
        """
        Generates a dictionary containing the basic info of the read data
        :return:
        """
        self.info = {}

    def global_coords_to_local(self, x_local_to_global, y_local_to_global):
        """Converts global data coordinates into local"""
        assert len(self.data.shape) >= 2 and self.data.shape[1] >= 2, "Error in data shape"
        self.data[:, 0] -= x_local_to_global
        self.data[:, 1] -= y_local_to_global

    def local_coords_to_global(self, x_local_to_global, y_local_to_global):
        """Converts local data coordinates into global"""
        assert len(self.data.shape) >= 2 and self.data.shape[1] >= 2, "Error in data shape"
        self.data[:, 0] += x_local_to_global
        self.data[:, 1] += y_local_to_global

    def dump_to_csv(self, output_file, delimiter=","):
        """
        Writes the data into a csv file
        :param output_file:
        :return:
        """
        print(f"Starting dump into {output_file}")
        np.savetxt(output_file, self.get_data(), delimiter=delimiter)
        print(f"The data has been properly exported to the {output_file} file")

    def create_postprocess_dict(self):
        self.postprocessing_dict = Path().cwd() / "postprocess"
        self.postprocessing_dict.mkdir(exist_ok=True)

    def generate_subfish_data(self, subfish_dict: dict, unit_factor=1/(365 * 24), unit_name='d') -> pd.DataFrame:
        """
        This method reads a given subfish dict and returns the calculated results
        Args:
            subfish_dict: dictionary containing subfish module parameters
            unit_factor: factor multiplyin the results (originally computed in seconds)
        Returns:
            Array containing the times and computed values
        """
        if not subfish_dict:
            raise AttributeError('Please, provide a suitable subfish dict object')
        tang_sol = subfish.calculate_tang(subfish_dict)
        tang_sol[0] *= unit_factor
        tang_sol_pd = pd.DataFrame({f'Time [{unit_name}]': tang_sol[0],
                                    'Result [M]': tang_sol[1]
                                    }
                                   )
        return tang_sol_pd



"""
Centroid file reader
"""
import numpy as np
from .BaseReader import BaseReader
from pydelling.config import config
import logging
import pandas as pd
from pathlib import Path
logger = logging.getLogger(__name__)


class StructuredListReader(BaseReader):
    data: pd.DataFrame
    def __init__(self, filename=None, var_pos=3, var_name="var", var_type=np.float32, centroid_pos=(0, 3), header=False):
        self.var_pos = None
        self.var = None
        self.var_name = None
        self.var_type = None
        self.centroid_pos = None
        self.header = None
        self.filename = Path(filename) if filename else Path(config.structured_list_reader.filename)
        self.data: pd.DataFrame
        super().__init__(self.filename, var_pos=var_pos,
                         var_name=var_name,
                         var_type=var_type,
                         centroid_pos=centroid_pos,
                         header=header)

    def open_file(self, filename):
        self.read_file()

    def read_file(self):
        logger.info(f"Reading list file from {self.filename.stem}")
        _temp_array = []
        with open(config.structured_list_reader.filename, "r") as reading_file:
            for _ in range(config.structured_list_reader.header_offset):
                _read = reading_file.readline()
            for nz in range(config.structured_list_reader.nz):
                for ny in range(config.structured_list_reader.ny):
                    for nx in range(config.structured_list_reader.nx):
                        read_value = reading_file.readline().split()[0]
                        p = np.array([config.structured_list_reader.ox + nx * config.structured_list_reader.dx,
                                      config.structured_list_reader.oy + ny * config.structured_list_reader.dy,
                                      config.structured_list_reader.oz + nz * config.structured_list_reader.dz,
                                      ])  # Position vector
                        _temp_array.append(np.append(p, read_value))
                        # idx = nx + config.structured_list_reader.nx * ny + config.structured_list_reader.nx * config.structured_list_reader.ny * nz + config.structured_list_reader.header_offset
                        # _temp_array.append(np.array())
        grain_array = pd.DataFrame(_temp_array, columns=["x", "y", "z", "v"])
        # grain_array[grain_array["v"] < config.structured_list_reader.min_value] = config.structured_list_reader.min_value
        self.data = grain_array

    def read_header(self):
        """
        TODO: Add the header reader of the centroid file
        Reads the header of the file
        :return:
        """
        pass


    def get_data(self) -> np.ndarray:
        """
        Outputs the data
        :return: np.ndarray object containing centroid information and variable output
        """
        return self.data.values.astype(np.float)

    @property
    def coordinates(self):
        return self.data[["x", "y", "x"]].values.astype(np.float)

    @property
    def values(self):
        return self.data[["v"]].values.astype(np.float)

    def build_info(self):
        """
        Generates a dictionary containing the basic info of the read data
        :return:
        """
        self.info["reader"] = {"n_cells": self.data.shape[0],
                     "filename": self.filename,
                     "var_name": self.var_name,
                     "var_position": self.var_pos}

    def dump_to_csv(self, output_file, delimiter=","):
        """
        Writes the data into a csv file
        :param output_file:
        :return:
        """
        print(f"Starting dump into {output_file}")
        np.savetxt(output_file, self.get_data(), delimiter=delimiter)
        print(f"The data has been properly exported to the {output_file} file")




import numpy as np
import shapely.geometry as geom
import sympy as sp
from pydelling.utils.geometry import Plane, Segment, Point, Line
from typing import List
from .Fracture import Fracture


class PolygonFracture:
    local_id = 0
    eps = 1e-8
    def __init__(self, dip, dip_dir, x, y, z, size, aperture=0.01):
        # super().__init__(dip, dip_dir, x, y, z, size, aperture)
        self.side_points = None
        self.dip = dip
        self.dip_dir = dip_dir
        self.x_centroid = x
        self.y_centroid = y
        self.z_centroid = z
        self.size = size
        self.intersection_dictionary = {}
        self.aperture = aperture
        self.local_id = PolygonFracture.local_id
        PolygonFracture.local_id += 1

    def get_side_points_v1(self):
        """
        Finds the side points of the fracture
        Returns: Coordinates of the plane side points
        """
        phi = self.dip_dir / 360 * (2 * np.pi)
        theta = self.dip / 360 * (2 * np.pi)

        u = np.array([np.cos(theta) * np.sin(phi), np.cos(theta) * np.cos(phi), -np.sin(theta)])
        v = np.array([np.cos(phi), - np.sin(phi), 0])

        A = self.centroid + self.size / 2 * (u + v)
        B = self.centroid + self.size / 2 * (u - v)
        C = self.centroid - self.size / 2 * (u + v)
        D = self.centroid - self.size / 2 * (u - v)


        self.side_points = 4

        return np.array([A, B, C, D])


    def get_side_points_v3(self):
        """
        alpha = 45; %strike
        delta = 30; %dip
        w = 4; %width (dipping side)
        L = 9; %Length (side parallel to the surface)

        A = [0,0,0]; % Pivot point at the surface

        H = w*sind(alpha+90);
        V = w*cosd(alpha+90);
        Z = -w*sind(delta);

        B = [L*sind(alpha) L*cosd(alpha) 0]+A;
        C = [L*cosd(90-alpha)+H L*sind(90-alpha)+V Z]+A;
        D = [w*sind(alpha+90) w*cosd(alpha+90) -w*sind(delta)]+A;

        P = [A; B; C; D];
        """
        alpha = self.dip_dir / 360 * (2 * np.pi)
        delta = self.dip / 360 * (2 * np.pi)
        w = self.size
        L = self.size

        A = np.array(self.centroid)
        A[0] = A[0] - np.sin(alpha) * L / 2
        A[1] = A[1] + np.cos(alpha) * L / 2
        A[2] = A[2] + np.sin(delta) * L / 2


        H = w * np.sin(alpha + np.pi / 2)
        V = w * np.cos(alpha + np.pi / 2)
        Z = -w * np.sin(delta)

        B = np.array([L * np.sin(alpha), L * np.cos(alpha), 0]) + A
        C = np.array([L * np.cos(np.pi / 2 - alpha) + H, L * np.sin(np.pi / 2 - alpha) + V, Z]) + A
        D = np.array([w * np.sin(alpha + np.pi / 2), w * np.cos(alpha + np.pi / 2), -w * np.sin(delta)]) + A

        P = np.array([A, B, C, D])

        return P

    def get_side_points_v2(self):
        """
        Finds the side points of the fracture
        Returns: Coordinates of the plane side points
        """
        alpha = self.dip / 360 * (2 * np.pi)
        beta = self.dip_dir / 360 * (2 * np.pi)

        A = self.centroid + np.array([
            + self.size / 2 * (-np.cos(beta) - np.sin(beta) * np.cos(alpha)),
            + self.size / 2 * (np.sin(beta) - np.cos(beta) * np.cos(alpha)),
            self.size / 2 * np.sin(alpha)
        ])

        B = self.centroid + np.array([
            + self.size / 2 * (-np.cos(beta) + np.sin(beta) * np.cos(alpha)),
            + self.size / 2 * (np.sin(beta) + np.cos(beta) * np.cos(alpha)),
            - self.size / 2 * np.sin(alpha)
        ])

        C = self.centroid + np.array([
            + self.size / 2 * (np.cos(beta) + np.sin(beta) * np.cos(alpha)),
            + self.size / 2 * (-np.sin(beta) + np.cos(beta) * np.cos(alpha)),
            - self.size / 2 * np.sin(alpha)
        ])

        D = self.centroid + np.array([
            + self.size / 2 * (np.cos(beta) - np.sin(beta) * np.cos(alpha)),
            + self.size / 2 * (-np.sin(beta) - np.cos(beta) * np.cos(alpha)),
            + self.size / 2 * np.sin(alpha)
        ])

        self.side_points = 4

        return np.array([A, B, C, D])


    def get_side_points(self, method='v1'):
        if method == 'v1':
            return self.get_side_points_v1()
        elif method == 'v2':
            return self.get_side_points_v2()
        elif method == 'v3':
            return self.get_side_points_v3()


    def to_obj(self, global_id=0, method='v1'):
        """Converts the fracture to an obj file"""
        side_points = self.get_side_points(method=method)
        obj_string = ''
        for i in range(len(side_points)):
            obj_string += 'v ' + str(side_points[i][0]) + ' ' + str(side_points[i][1]) + ' ' + str(side_points[i][2]) + '\n'
        obj_string += f'f '
        for i in range(len(side_points)):
            obj_string += str(global_id + i) + ' '
        obj_string += '\n'
        return obj_string

    @property
    def unit_normal_vector(self):
        """Returns the normal vector of the fracture"""
        get_side_points = self.get_side_points()
        v1 = get_side_points[1] - get_side_points[0]
        v2 = get_side_points[2] - get_side_points[0]
        cross = np.cross(v1, v2)
        return cross / np.linalg.norm(cross)

    def distance_to_point(self, point: np.ndarray):
        """Returns the distance to a point"""
        distance_vector = self.centroid - point
        return np.dot(distance_vector, self.unit_normal_vector)

    def get_bounding_box(self):
        """Returns the bounding box of the fracture"""
        side_points = self.get_side_points()
        x_min = np.min(side_points[:, 0])
        x_max = np.max(side_points[:, 0])
        y_min = np.min(side_points[:, 1])
        y_max = np.max(side_points[:, 1])
        z_min = np.min(side_points[:, 2])
        z_max = np.max(side_points[:, 2])
        return np.array([x_min, x_max, y_min, y_max, z_min, z_max])

    def point_inside_bounding_box(self, point: np.ndarray):
        """Returns if a point is inside the bounding box of the fracture"""
        bounding_box = self.get_bounding_box()
        if point[0] < bounding_box[0] or point[0] > bounding_box[1]:
            return False
        elif point[1] < bounding_box[2] or point[1] > bounding_box[3]:
            return False
        elif point[2] < bounding_box[4] or point[2] > bounding_box[5]:
            return False
        else:
            return True

    @property
    def centroid(self):
        return np.array([self.x_centroid, self.y_centroid, self.z_centroid])

    @property
    def polygon(self):
        side_points = self.get_side_points()
        self._polygon = geom.Polygon(side_points)
        return self._polygon

    @property
    def sympy_plane(self):
        """Returns the plane of the fracture"""
        return sp.Plane(self.centroid, normal_vector=self.unit_normal_vector)

    @property
    def plane(self):
        return Plane(self.centroid, normal=self.unit_normal_vector)

    @property
    def corners(self) -> List[Point]:
        """Returns the corners of the fracture"""
        return [Point(point) for point in self.get_side_points()]

    @property
    def corner_segments(self):
        """Returns the corner lines of the fracture"""
        corner_segments = [
            Segment(self.corners[0], self.corners[1]),
            Segment(self.corners[1], self.corners[2]),
            Segment(self.corners[2], self.corners[3]),
            Segment(self.corners[3], self.corners[0])
        ]
        return corner_segments

    @property
    def corner_lines(self):
        """Returns the corner lines of the fracture"""
        corner_segments = [
            Line(self.corners[0], self.corners[1]),
            Line(self.corners[1], self.corners[2]),
            Line(self.corners[2], self.corners[3]),
            Line(self.corners[3], self.corners[0])
        ]
        return corner_segments

    def contains(self, point: Point):
        """Returns if a point is inside the fracture"""
        q1, q2, q3, q4 = self.corners
        q1: Point

        largest_normal_index = self.largest_index_normal_vector
        q1_hat = np.delete(q1, largest_normal_index)
        q2_hat = np.delete(q2, largest_normal_index)
        q3_hat = np.delete(q3, largest_normal_index)
        q4_hat = np.delete(q4, largest_normal_index)
        p_hat = np.delete(point, largest_normal_index)
        u0 = p_hat[0]
        u1 = q1_hat[0]
        u2 = q2_hat[0]
        u3 = q3_hat[0]
        u4 = q4_hat[0]
        v0 = p_hat[1]
        v1 = q1_hat[1]
        v2 = q2_hat[1]
        v3 = q3_hat[1]
        v4 = q4_hat[1]


        s1 = (v1 - v2) * u0 + (u2 - u1) * v0 + v2 * u1 - u2 * v1
        s2 = (v2 - v3) * u0 + (u3 - u2) * v0 + v3 * u2 - u3 * v2
        s3 = (v3 - v4) * u0 + (u4 - u3) * v0 + v4 * u3 - u4 * v3
        s4 = (v4 - v1) * u0 + (u1 - u4) * v0 + v1 * u4 - u1 * v4

        s = np.array([s1, s2, s3, s4])

        equal_sign = np.all(s >= -self.eps) if s[0] >= -self.eps else np.all(s <= self.eps)

        if equal_sign:
            return True
        else:
            return False

    @property
    def largest_index_normal_vector(self):
        """Returns the largest coordinate index of the normal vector"""
        return np.argmax(self.unit_normal_vector)









import numpy as np

from .Fracture import Fracture
import meshio
import trimesh
from trimesh.exchange.misc import load_meshio
import trimesh.proximity as proximity
from pathlib import Path
from tqdm import tqdm


class Fault:
    local_id = 0
    def __init__(self, filename=None,
                 mesh=None,
                 polygon=None,
                 aperture=None,
                 transmissivity=None,
                 porosity=None,
                 storativity=None,
                 ):

        if filename is not None:
            self.meshio_mesh = meshio.read(filename)
        if mesh is not None:
            self.meshio_mesh: meshio.Mesh = mesh
        self.trimesh_mesh: trimesh.Trimesh = trimesh.load_mesh(filename)
        self.aperture = aperture
        self.associated_elements = []
        self.transmissivity = transmissivity
        self.porosity = porosity
        self.storativity = storativity

        Fault.local_id += 1

    def distance(self, points: np.ndarray, n_max: int = 2500):
        if points.shape[0] == 3:
            points = points.reshape(1, 3)
        # Divide the points into chunks of n_max
        n_chunks = int(points.shape[0] / n_max)
        if n_chunks == 0:
            n_chunks = 1
        distances = []
        for i in tqdm(range(n_chunks), desc=f"Computing {len(points)} distances"):
            distances.append(proximity.signed_distance(self.trimesh_mesh, points[i * n_max:(i + 1) * n_max]))
        return np.concatenate(distances)

    def _to_obj(self, global_id=0):
        """Converts the fault to an obj file"""
        str_obj = ""
        for i, f in enumerate(self.meshio_mesh.points):
            str_obj += f"v {f[0]} {f[1]} {f[2]}\n"

        for i, f in enumerate(self.meshio_mesh.cells[0].data):
            str_obj += "f "
            for j in range(3):
                str_obj += str(f[j] + global_id) + " "
            str_obj += "\n"
        return str_obj

    def to_obj(self, filename=None, global_id=1):
        str_obj = self._to_obj(global_id=global_id)
        if filename is not None:
            with open(filename, 'w') as f:
                f.write(str_obj)
        return str_obj


    @property
    def points(self):
        return self.meshio_mesh.points

    @property
    def cells(self):
        return self.meshio_mesh.cells[0].data

    @property
    def num_points(self):
        return self.meshio_mesh.points.shape[0]

    @property
    def num_cells(self):
        return self.meshio_mesh.cells[0].data.shape[0]

    @property
    def centroid(self):
        return self.trimesh_mesh.centroid

    @property
    def size(self):
        return np.sqrt(self.trimesh_mesh.area)

    @property
    def normal_vector(self):
        return np.mean(self.trimesh_mesh.face_normals, axis=0)











"""
Base interface for a reader class
"""
import numpy as np
import logging
from pydelling.readers import BaseReader
logger = logging.getLogger(__name__)
from pydelling.config import config
import logging
from pathlib import Path
import h5py
import natsort
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
logger = logging.getLogger(__name__)
from collections import OrderedDict


class PflotranReader(BaseReader):
    def __init__(self, filename=None):
        self.filename = Path(filename) if filename else Path(config.pflotran_reader.filename)
        logger.info(f"Reading PFLOTRAN results file from {self.filename}")
        super().__init__(filename=self.filename)
        self.results = {}
        for time in self.time_keys:
            self.results[time] = PflotranResults(time=time, data=self.data[self.time_keys[time]])
        self.variables = list(self.results[self.time_values[0]].variable_keys)

    def open_file(self, filename):
        self.data: h5py.File = h5py.File(self.filename, 'r')
        # Obtain time keys
        self.time_dict_keys = natsort.natsorted(OrderedDict({float(key.split(' ')[2]): key for key in self.data.keys() if 'Time' in key}))
        self.time_keys = OrderedDict({float(key.split(' ')[2]): key for key in self.data.keys() if 'Time' in key})
        self.time_keys = {key: self.time_keys[key] for key in self.time_dict_keys}

    @property
    def time_values(self):
        return natsort.natsorted(self.time_keys.keys())

    @property
    def coordinates(self):
        temp_coordinates = self.data['Coordinates']
        temp_df = {'x[m]': np.array(temp_coordinates['X [m]']),
                                'y[m]': np.array(temp_coordinates['Y [m]']),
                                'z[m]': np.array(temp_coordinates['Z [m]']),
                                }
        return temp_df

    def get_data(self) -> np.ndarray:
        """
        Outputs the read data
        :return:
        """
        return np.array(0)

    def build_info(self):
        """
        Generates a dictionary containing the basic info of the read data
        :return:
        """
        self.info = {}

    def get_observation_point(self, variable, point=(0, 0, 0)):
        """
        Gets the time dependent evolution of a given variable at a given mesh point
        Args:
            variable: variable name to extract the data from
            point: mesh location

        Returns:
            Array containing the temporal evolution of the variable
        """
        temp_array = []
        for time in self.time_values:
            temp_array.append(self.results[time].results[variable][point[0], point[1], point[2]])
        pd_array = pd.DataFrame(np.array(temp_array), columns=[variable])
        return pd_array

    @property
    def mineral_names(self):
        temp_keys = [key.split('_')[0] for key in self.variables if 'VF' in key]
        return temp_keys

    @property
    def species_names(self):
        temp_keys = [key.split('_')[1] for key in self.variables if 'Total' in key]
        return temp_keys

    @property
    def x_centroid(self):
        return pd.DataFrame(np.diff(self.coordinates['x[m]']) + self.coordinates['x[m]'][0:-1], columns=['x[m]'])

    @property
    def y_centroid(self):
        return np.diff(self.coordinates['y[m]']) + self.coordinates['y[m]'][0:-1]

    @property
    def z_centroid(self):
        return np.diff(self.coordinates['z[m]']) + self.coordinates['z[m]'][0:-1]


    def get_mineral_vf_key(self, mineral):
        """
        Returns the correct key of the mineral volume fraction name
        Args:
            mineral: mineral name

        Returns:
            mineral volume fraction key
        """
        return f"{mineral}_VF [m^3 mnrl_m^3 bulk]"

    def get_mineral_rate_key(self, mineral):
        """
        Returns the correct key of the mineral rate name
        Args:
            mineral: mineral name

        Returns:
            mineral rate key
        """
        return f"{mineral}_Rate [mol_m^3_sec]"

    def get_mineral_si_key(self, mineral):
        """
        Returns the correct key of the mineral si name
        Args:
            mineral: mineral name

        Returns:
            mineral si key
        """
        return f"{mineral}_SI"

    def get_primary_species_key(self, species):
        """
        Returns the correct key of the given species name
        Args:
            species: specie name

        Returns:
            specie key
        """
        return f"Total_{species}"

    def plot_primary_species(self, type='1D', postprocess_dir='./postprocess'):
        """This method plots the primary species of a 1D model"""
        logger.info(f'Plotting [{self.species_names}] primary species from {self.filename}')
        postprocess_dir = Path(postprocess_dir)
        if type == '1D':
            for species in self.species_names:
                # Generate rate plots
                primary_species_key = self.get_primary_species_key (species)
                plot_df = pd.DataFrame()
                plt.clf()
                for time in config.postprocessing.times:
                    specie_data = self.results[time].results[primary_species_key][:, 0, 0]
                    specie_data_pd = pd.DataFrame (specie_data, columns=[f'{time} years'])
                    plot_df = plot_df.combine_first (specie_data_pd)
                plot_df = plot_df.combine_first (self.x_centroid)
                plot_df = plot_df.set_index ('x[m]')
                plot_df = plot_df.reindex(natsort.natsorted(plot_df.columns), axis=1)
                line_plot: plt.Axes = sns.lineplot (data=plot_df)
                line_plot.set_xlabel ('X [m]')
                line_plot.set_ylabel (f'{species}')
                plt.savefig(postprocess_dir / f'{species}.png')

    def plot_vf_variation(self, type='1D', postprocess_dir='./postprocess', ignored_minerals=[]):
        """This method plots the vf variation due to mineral precipitation"""
        logger.info(f'Plotting [{self.mineral_names}] primary species from {self.filename}')
        postprocess_dir = Path(postprocess_dir)
        postprocess_dir.mkdir(exist_ok=True)
        if type == '1D':
            for mineral in self.mineral_names:
                if mineral in ignored_minerals:
                    continue
                # Generate porosity variation plots
                primary_minerals_key = self.get_mineral_vf_key(mineral)
                plot_df = pd.DataFrame()
                plt.clf()
                # Save initial mineral volumes and porosity
                assert 'Porosity' in self.results[0.0].results, 'Porosity must be within the PFLOTRAN output variables'
                initial_porosity = self.results[0.0].results['Porosity']
                initial_mineral_vf = self.results[0.0].results[primary_minerals_key][:, 0, 0]
                for time in config.postprocessing.times:
                    mineral_vf = self.results[time].results[primary_minerals_key][:, 0, 0]
                    mineral_variation = mineral_vf - initial_mineral_vf
                    mineral_vf_pd = pd.DataFrame (mineral_variation, columns=[f'{time} years'])
                    plot_df = plot_df.combine_first (mineral_vf_pd)
                plot_df = plot_df.combine_first (self.x_centroid)
                plot_df = plot_df.set_index ('x[m]')
                plot_df = plot_df.reindex(natsort.natsorted(plot_df.columns), axis=1)
                line_plot: plt.Axes = sns.lineplot (data=plot_df)
                line_plot.set_xlabel ('X [m]')
                line_plot.set_ylabel (f'{mineral} volume fraction variation')
                plt.savefig(postprocess_dir / f'{mineral}.png')

    def plot_total_porosity_variation(self, type='1D', postprocess_dir='./postprocess', ignored_minerals=[]):
        """This method plots the vf variation due to mineral precipitation"""
        logger.info(f'Plotting [{self.mineral_names}] primary species from {self.filename}')
        postprocess_dir = Path(postprocess_dir)
        postprocess_dir.mkdir(exist_ok=True)
        plot_df = pd.DataFrame()

        if type == '1D':
            assert 'Porosity' in self.results[0.0].results, 'Porosity must be within the PFLOTRAN output variables'
            initial_porosity = self.results[0.0].results['Porosity'][:, 0, 0]
            for time in config.postprocessing.times:
                total_porosity_variation = np.zeros_like(initial_porosity)
                for mineral in self.mineral_names:
                    if mineral in ignored_minerals:
                        continue
                    # Generate porosity variation plots
                    primary_minerals_key = self.get_mineral_vf_key(mineral)
                    initial_mineral_vf = self.results[0.0].results[primary_minerals_key][:, 0, 0]
                    mineral_vf = self.results[time].results[primary_minerals_key][:, 0, 0]
                    mineral_variation = mineral_vf - initial_mineral_vf
                    total_porosity_variation += mineral_variation

                total_porosity_variation = total_porosity_variation / initial_porosity * 100
                mineral_vf_pd = pd.DataFrame(total_porosity_variation, columns=[f'{time} years'])
                plot_df = plot_df.combine_first(mineral_vf_pd)
            plot_df = plot_df.combine_first(self.x_centroid)
            plot_df = plot_df.set_index('x[m]')
            plot_df = plot_df.reindex(natsort.natsorted(plot_df.columns), axis=1)
            line_plot: plt.Axes = sns.lineplot (data=plot_df)
            line_plot.set_xlabel ('X [m]')
            line_plot.set_ylabel (f'{mineral} volume fraction variation')
        plt.savefig(postprocess_dir / f'{mineral}.png')

class PflotranResults:
    """
    This class stores the PFLOTRAN results
    """
    def __init__(self, time, data):
        """
        Args:
            time: timestep
            data: results dataset
        """
        self.time = time
        self.raw_data = data
        self.results = {key: np.array(self.raw_data[key]) for key in self.raw_data}
        self.variable_keys = self.results.keys()

    def __repr__(self):
        return f"Results of time {self.time}"

    @property
    def mineral_names(self):
        temp_keys = [key.split('_')[0] for key in self.variable_keys if 'VF' in key]
        return temp_keys

    @property
    def species_names(self):
        temp_keys = [key.split('_')[1] for key in self.variable_keys if 'Total' in key]
        return temp_keys








from pydelling.paraview_processor.filters import BaseFilter
try:
    from paraview.simple import *
    from paraview.vtk.numpy_interface import dataset_adapter as dsa
    from paraview.vtk.numpy_interface import algorithms as algs
    from paraview import servermanager as sm
except:
    pass


import pandas as pd
import numpy as np


class PlotOverLineFilter(BaseFilter):
    """
    This class implements the PlotOverLine paraview filter
    """
    filter_type: str = "PlotOverLineFilter"
    counter: int = 0

    def __init__(self, input_filter, name, point_1=None, point_2=None, n_line=None):
        super().__init__(name=name)
        PlotOverLineFilter.counter += 1
        self.filter = PlotOverLine(Input=input_filter)
        if point_1:
            assert point_2, "Two points need to be defined"
        if point_2:
            assert point_1, "Two points need to be defined"
        if point_1 or point_2:
            self.filter.Point1 = point_1
            self.filter.Point2 = point_2
        if n_line:
            self.filter.Resolution = n_line

    def set_points(self, point_1, point_2):
        """
        This method sets the `point_1` and `point_2` variables to the filter
        Args:
            point_1: First point of the line
            point_2: Second point of the line
        """
        self.filter.Point1 = point_1
        self.filter.Point2 = point_2

    def set_line_resolution(self, n):
        """
        This methods modifies the resolution of the line used to interpolate the data
        Args:
            n: Number specifying the number of divisions of the line used to interpolate the data on.
        """
        self.filter.Resolution = n

    @property
    def point_data(self):
        vtk_object = sm.Fetch(self.filter)
        vtk_object = dsa.WrapDataObject(vtk_object)
        pd_df = pd.DataFrame()
        for key in self.point_keys:
            temp_dataset = np.array(vtk_object.PointData[key]).transpose()
            if len(temp_dataset.shape) != 1:
                # The dataset is a vector:
                for idx, vector_element in enumerate(temp_dataset):
                    new_key = f"{key}{self.vector_keys[idx]}"
                    pd_df[new_key] = vector_element
            else:
                pd_df[key] = temp_dataset
        return pd_df.dropna()





import unittest
from pydelling.preprocessing import DfnPreprocessor


class DfnPreprocessorCase(unittest.TestCase):
    def test_dfn_sum(self):
        dfn_one = DfnPreprocessor()
        dfn_one.add_fracture(
            x=0, y=0, z=0, dip=0, dip_dir=0, aperture=0.1
        )
        dfn_two = DfnPreprocessor()
        dfn_two.add_fracture(
            x=0, y=0, z=0, dip=0, dip_dir=0, aperture=0.5
        )
        dfn_three = dfn_one + dfn_two
        self.assertEqual(len(dfn_three.dfn), 2)


if __name__ == '__main__':
    unittest.main()


from . import Line, Point, Plane, Segment
import numpy as np

def intersect_line_line(line_1: Line, line_2: Line):
    if line_1.is_parallel(line_2):
        return None
    else:
        delta_p = line_1.p - line_2.p
        a = np.array([
            [line_1.direction_vector[0], -line_2.direction_vector[0]],
            [line_1.direction_vector[1], -line_2.direction_vector[1]],
            [line_1.direction_vector[2], -line_2.direction_vector[2]]

        ])
        b = - np.array([
            [delta_p[0]],
            [delta_p[1]],
            [delta_p[2]],
        ])
        x = np.linalg.lstsq(a, b, rcond=None)
        if x[1] >= line_1.eps:
            return None
        else:
            return Point(line_1.p + line_1.direction_vector * x[0][0])

def intersect_plane_plane(plane_1: Plane, plane_2: Plane):
    """Performs the intersection of this plane with the given plane"""
    if plane_1.is_parallel(plane_2):
        return None
    normal_a = plane_1.n
    normal_b = plane_2.n
    d_a = np.dot(plane_1.p, plane_1.n)
    d_b = np.dot(plane_2.p, plane_2.n)
    U = np.cross(normal_a, normal_b)
    M = np.array((normal_a, normal_b, U))
    X = np.array([d_a, d_b, 0.0]).reshape(3, 1)
    # print(M)
    p_inter = np.linalg.solve(M, X).T
    p1 = p_inter[0]
    p2 = (p_inter + U)[0]
    intersected_line = Line(p1, p2)

    return intersected_line

def intersect_plane_line(plane: Plane, line: Line):
    """Performs the intersection of this plane with the given line"""
    # The point r = q + lambda * v
    dot_n_diff = np.dot(plane.n, plane.p - line.p)
    dot_n_v = np.dot(plane.n, line.direction_vector)
    if dot_n_v == 0: # Plane and line are parallel
        return None
    else:
        lambda_ = dot_n_diff / dot_n_v
        return Point(line.p + lambda_ * line.direction_vector)

def intersect_plane_segment(plane: Plane, segment: Segment):
    """Performs the intersection of this plane with the given segment"""
    # The point r = q + lambda * v
    dot_n_diff = np.dot(plane.n, plane.p - segment.p1)
    dot_n_v = np.dot(plane.n, segment.direction_vector)
    if dot_n_v == 0: # Plane and line are parallel
        return None
    else:
        lambda_ = dot_n_diff / dot_n_v
        intersection_point = Point(segment.p1 + lambda_ * segment.direction_vector)
        if segment.contains(intersection_point):
            return intersection_point
        else:
            return None








import pandas as pd

from pydelling.webapps import WebAppRunner
from pydelling.webapps.components import InputComponent
import streamlit as st
from pydelling.config import config
from pydelling.interpolation import SparseDataInterpolator
import numpy as np
import streamlit.components.v1 as components


class SparseInterpolatorWebapp(WebAppRunner):
    def construct(self):
        if 'input_selection_done' not in st.session_state:
            st.session_state['input_selection_done'] = False
        if 'mesh_input_done' not in st.session_state:
            st.session_state['mesh_input_done'] = False
        if 'interpolation_done' not in st.session_state:
            st.session_state['interpolation_done'] = False

        st.header('Sparse data interpolator')
        st.markdown("""
        This webapp allows you to upload sparse data (A cloud of points) in `|x|y|z|value|` format and interpolate it 
        into a mesh or another target cloud of points.

        These are the steps:

        """)
        st.subheader('Step 1: Upload sparse data')
        InputComponent(input_type='csv', key='input_data')
        if st.session_state['input_data_done']:
            self.input_data: pd.DataFrame = st.session_state['input_data']

            with st.form('select input variables'):
                st.markdown("""Select the columns corresponding to the`x`, `y` and `z` variables and the `target variable`""")
                cols = st.columns(4)
                x_index = None
                y_index = None
                z_index = None
                for col_idx, column_name in enumerate(self.input_data.columns):
                    if 'x' == column_name.lower():
                        x_index = col_idx
                    elif 'y' == column_name.lower():
                        y_index = col_idx
                    elif 'z' == column_name.lower():
                        z_index = col_idx

                with cols[0]:
                    x_col = st.selectbox('x centroid', self.input_data.columns, index=x_index if x_index is not None else 0)
                with cols[1]:
                    y_col = st.selectbox('y centroid', self.input_data.columns, index=y_index if y_index is not None else 0)
                with cols[2]:
                    z_col = st.selectbox('z centroid', self.input_data.columns, index=z_index if z_index is not None else 0)
                with cols[3]:
                    target_col = st.selectbox('target variable', self.input_data.columns, index=len(self.input_data.columns) - 1)
                submit = st.form_submit_button('Select variables')
                if submit:
                    st.success(f'centroid variables: [{x_col}, {y_col}, {z_col}] and target variable: {target_col} selected')
                    st.session_state['input_selection_done'] = True

        if st.session_state['input_selection_done']:
            st.subheader('Step 2: Upload mesh data')
            selected_type = st.selectbox('Select target mesh file format', config.globals.webapps.supported_mesh_data)
            if selected_type == 'csv':
                InputComponent(input_type='csv', key='mesh_input')
                with st.form('select input variables mesh'):
                    st.markdown(
                        """Select the columns corresponding to the`x`, `y` and `z` variables of the centroids""")
                    cols = st.columns(3)
                    x_index = None
                    y_index = None
                    z_index = None
                    for col_idx, column_name in enumerate(self.input_data.columns):
                        if 'x' == column_name.lower():
                            x_index = col_idx
                        elif 'y' == column_name.lower():
                            y_index = col_idx
                        elif 'z' == column_name.lower():
                            z_index = col_idx

                    with cols[0]:
                        x_col = st.selectbox('x centroid', self.input_data.columns,
                                             index=x_index if x_index is not None else 0)
                    with cols[1]:
                        y_col = st.selectbox('y centroid', self.input_data.columns,
                                             index=y_index if y_index is not None else 0)
                    with cols[2]:
                        z_col = st.selectbox('z centroid', self.input_data.columns,
                                             index=z_index if z_index is not None else 0)
                    submit = st.form_submit_button('Select variables')
                    if submit:
                        st.success(
                            f'centroid variables: [{x_col}, {y_col}, {z_col}] selected as the target mesh')
                        st.session_state['input_selection_done'] = True

            elif selected_type == 'vtk':
                InputComponent(input_type='vtk', key='mesh_input')

            elif selected_type == 'fem':
                InputComponent(input_type='fem', key='mesh_input')

        if st.session_state['mesh_input_done']:
            st.subheader('Step 3: Interpolate data')
            input_data = self.input_data[[x_col, y_col, z_col, target_col]]
            if selected_type == 'vtk':
                mesh_data = st.session_state['mesh_input']
                mesh_data = mesh_data.centroids
            elif selected_type == 'csv':
                mesh_data = st.session_state['mesh_input'][[x_col, y_col, z_col]].values
            elif selected_type == 'fem':
                mesh_data = st.session_state['mesh_input'].centroids
            with st.form('interpolation_step'):
                st.markdown("""Choose the interpolation method""")
                interpolation_method = st.selectbox('interpolation method', ['nearest', 'linear', 'cubic'], index=0)

                interpolator = SparseDataInterpolator(interpolation_data=input_data, mesh_data=mesh_data)
                submit = st.form_submit_button()
                if submit:
                    with st.spinner('interpolating data'):
                        interpolator.run(method=interpolation_method)
                    st.success(f'interpolation done')
                    interpolated_data = interpolator.interpolated_data
                    interpolated_block = np.array([[mesh_data[i, 0], mesh_data[i, 1], mesh_data[i, 2], interpolated_data[i]] for i in range(len(mesh_data))])
                    interpolated_block = pd.DataFrame(interpolated_block, columns=['x', 'y', 'z', target_col])
                    st.session_state['interpolation_done'] = True
                    st.session_state['interpolated_data'] = interpolated_block
        if st.session_state['interpolation_done']:
            interpolated_block = st.session_state['interpolated_data']

            st.subheader('Step 4: Download interpolated data')
            st.download_button(
                label="Download data as CSV",
                data=self.convert_df(interpolated_block),
                file_name='interpolated_data.csv',
                mime='text/csv',
            )
            visualize = st.checkbox('Visualize interpolated data')
            if visualize:
                import pyvista as pv
                interpolated_block = st.session_state['interpolated_data']
                pl = pv.Plotter()
                point_cloud = pv.PolyData(interpolated_block[['x', 'y', 'z']].values)
                point_cloud['value'] = interpolated_block[target_col].values
                point_size = point_cloud.bounds
                point_size = np.max(point_size) * 0.0075

                pl.add_mesh(point_cloud, point_size=point_size, style='points')
                pl.background_color = 'white'
                pl.add_scalar_bar(
                    'value',
                    interactive=True, vertical=False,
                    title_font_size=35,
                    label_font_size=30,
                    outline=True, fmt='%10.5f',
                )
                pl.export_html('pyvista.html')  # doctest:+SKIP

                HtmlFile = open("pyvista.html", 'r', encoding='utf-8')
                source_code = HtmlFile.read()
                components.html(source_code, height=750, width=1000)

    @staticmethod
    @st.cache
    def convert_df(df):
        return df.to_csv(index=False).encode('utf-8')


if __name__ == '__main__':
    webapp = SparseInterpolatorWebapp()
    webapp.run()


from pydelling.paraview_processor.filters import BaseFilter
try:
    from paraview.simple import *
except:
    pass


class CalculatorFilter(BaseFilter):
    filter_type: str = "Calculator"
    counter: int = 0

    def __init__(self, input_filter, function, name, output_array_name=None, attribute_type='Cell Data'):
        super().__init__(name=name)
        CalculatorFilter.counter += 1
        self.filter = Calculator(Input=input_filter)
        self.filter.Function = function
        self.filter.AttributeType = attribute_type
        if output_array_name:
            self.filter.ResultArrayName = output_array_name

    def set_attribute_type(self, attribute_type):
        self.filter.AttributeType = attribute_type

    def set_function(self, function):
        self.filter.Function = function

    @property
    def calculation(self):
        """
        This property returns the calculated array
        Returns:
            The array containing the calculation
        """
        if self.filter.AttributeType == "Cell Data":
            return self.cell_data[self.filter.ResultArrayName]
        if self.filter.AttributeType == "Point Data":
            return self.point_data[self.filter.ResultArrayName]




from .BaseReader import BaseReader
from .CentroidReader import CentroidReader
from .RasterFileReader import RasterFileReader
from .StructuredGridReader import StructuredGridReader
from .StructuredListReader import StructuredListReader
from .OpenFoamReader import OpenFoamReader
from .StreamlineReader import StreamlineReader
from .VtkReader import VtkReader
from .iGPReader.io.iGPReader import iGPReader
from .ConnectFlowReader import ConnectFlowReader
from .PflotranReader import PflotranReader
from .PflotranObservationPointReader import PflotranObservationPointReader
from .FemReader import FemReader
from .UnstructuredMeshReader import UnstructuredMeshReader
from .VTKMeshReader import VTKMeshReader
from .SmeshReader import SmeshReader


from pydelling.paraview_processor.filters import BaseFilter
try:
    from paraview.simple import *
except:
    pass


class StreamTracerWithCustomSourceFilter(BaseFilter):
    """
    This class implements the Stream Tracer with custom source filter
    """
    filter_type: str = "Stream_tracer_with_custom_source"
    counter: int = 0

    def __init__(self, input_filter, seed_source, name):
        super().__init__(name=name)
        StreamTracerWithCustomSourceFilter.counter += 1
        self.filter = StreamTracerWithCustomSource(Input=input_filter)
        self.filter.SeedSource = seed_source

    def set_seed_source(self, seed_source):
        self.filter.SeedSource = seed_source

    def add_vector(self, vector_name):
        self.filter.Vectors = ['POINTS', vector_name]

    def add_minimum_step_length(self, minimum_step_length):
        self.filter.MinimumStepLength = minimum_step_length


import unittest
from pydelling.interpolation import KdeEstimator
import numpy as np
import pandas as pd


class KdeEstimatorCase(unittest.TestCase):
    def setUp(self) -> None:
        self.test_data = pd.DataFrame(np.random.normal(size=1000))

    def test_set_up(self):
        self.kde_estimator = KdeEstimator()

    def test_train_gaussian(self):
        self.kde_estimator = KdeEstimator(data=self.test_data, bandwidth=0.0001)
        self.kde_estimator.run()
        test_samples = self.kde_estimator.sample(100)
        self.assertLess(abs(test_samples.mean()), 0.5)

if __name__ == '__main__':
    unittest.main()


from .BaseFilter import BaseFilter
from .CalculatorFilter import CalculatorFilter
from .IntegrateVariablesFilter import IntegrateVariablesFilter
from .PlotOverLineFilter import PlotOverLineFilter
from .VtkFilter import VtkFilter
from .XDMFFilter import XDMFFilter
from .CellDataToPointDataFilter import CellDataToPointDataFilter
from .ClipFilter import ClipFilter
from .StreamTracerWithCustomSourceFilter import StreamTracerWithCustomSourceFilter
from .SaveDataFilter import SaveDataFilter
from .AppendArcLengthFilter import AppendArcLengthFilter
from .TableToPointsFilter import TableToPointsFilter
from .CsvReaderFilter import CsvReaderFilter
from .TecplotFilter import TecplotFilter


import numpy as np
from typing import *


class BaseAbstractMeshObject:
    '''Class for the base abstract mesh object. Provides method and attribute namespace'''
    nodes: np.ndarray  # Node id set
    coords: np.ndarray   # Coordinates of each node
    local_id = int  # Element id
    type: str or None
    meshio_type: str or None
    centroid: np.ndarray  # Centroid of the element

    @property
    def edges(self):
        '''Returns the edges of the element'''
        return None

    @property
    def edge_vectors(self):
        '''Returns the edge vectors of the element'''
        return None




"""
Module containing general utility methods
"""

import pandas as pd
import numpy as np
from scipy.interpolate import griddata
import logging
import os
logger = logging.getLogger(__name__)


def get_config_path():
    from pathlib import Path
    return Path(__file__).parent.parent / "config"


def get_output_path():
    from pathlib import Path
    os.makedirs(Path.cwd() / "output", exist_ok=True)
    return Path.cwd() / "output"


def get_material_id_from_centroids(igp_reader, borehole_coordinates) -> np.array:
    from pydelling.readers.iGPReader.io import iGPReader, BoreholeReader
    igp_reader: iGPReader = igp_reader
    borehole_coordinates: pd.DataFrame = borehole_coordinates
    centroid_coordinates: np.array = igp_reader.centroids

    borehole_coordinates_values = borehole_coordinates.values[:, 0:3]  # Get numpy array of borehole coordinate values
    material_values = np.empty(shape=igp_reader.centroids.shape[0])
    for material_id, material in enumerate(igp_reader.material_dict):
        material_values[igp_reader.material_dict[material] - 1] = material_id

    interpolated_materials = griddata(centroid_coordinates, material_values, borehole_coordinates_values, method="nearest")
    return interpolated_materials



"""
Centroid file reader
"""
import numpy as np
from .BaseReader import BaseReader
from pydelling.config import config
from pydelling.utils.utils import aperture_from_a_xy_point
from pydelling.paraview_processor import ParaviewProcessor
import logging
import pandas as pd
import pickle
from pandas.core.groupby import DataFrameGroupBy
from tqdm import tqdm
import random

logger = logging.getLogger(__name__)


class StreamlineReader(BaseReader):
    data: pd.DataFrame
    raw_data: pd.DataFrame
    stream_data: DataFrameGroupBy
    is_aperture_zero: bool = False

    def __init__(self, filename, header=False):
        self.header = None
        super().__init__(filename,
                         header=header)

    def read_file(self, opened_file):
        """
        This method reads an opened file an sets the `self` variables properly to be used throuhgout the class
        Args:
            opened_file: the object containing the opened file
        """
        logger.info(f"Reading streamlines from {self.filename}")
        self.data = pd.read_csv(opened_file)
        self.raw_data = self.data.copy()  # Copy of original data
        self.data = self.data.rename(columns={"Points:0": "x",
                                              "Points:1": "y",
                                              "Points:2": "z",
                                              }
                                     )
        self.data = self.data.apply(pd.to_numeric, args=("coerce",))
        self.stream_data: DataFrameGroupBy = self.data.groupby("SeedIds")

    def compute_arrival_times(self,
                              reason_of_termination=None,
                              min_x=None,
                              min_y=None,
                              ) -> pd.Series:
        """
        This method computes the arrival times of the streamlines

        Args:
            reason_of_termination: Filter by the Paraview tag that defines the status of the streamlines
            min_x: Filters the streamlines with their max x < min_x

        Returns:
             A pd.Series object containing the arrival times of the streamlines

        """
        logger.info("Computing arrival times of the streamlines")
        filtered_streamlines = self.filter_streamlines(reason_of_termination=reason_of_termination, min_x=min_x, min_y=min_y)
        temp_series: pd.Series = filtered_streamlines.max()["IntegrationTime"]
        return temp_series

    def compute_arrival_times_per_material(self, reason_of_termination=None) -> pd.Series:
        """
        This method computes the arrival times of the streamlines for a particular material
        Returns:
             A pd.Series object containing the arrival times of the streamlines
        """
        logger.info("Computing arrival times of the streamlines per material")
        reason_of_termination = reason_of_termination if reason_of_termination else config.streamline_reader.reason_of_termination
        temp_df = self.stream_data
        if reason_of_termination:
            temp_df = temp_df.filter(lambda x: x["ReasonForTermination"].max() == reason_of_termination)

        temp_df['Material ID'] = (temp_df['Material ID'] + 0.45).apply(np.floor)
        temp_series: pd.Series = temp_df.groupby(["Material ID", "SeedIds"]).max()['IntegrationTime']
        temp_series = temp_series.reset_index()

        return temp_series



    def compute_arrival_times_per_material_paula(self, reason_of_termination=None) -> pd.Series:
        """
        This method computes the arrival times of the streamlines for a particular material
        Returns:
             A pd.Series object containing the arrival times of the streamlines
        """
        logger.info("Computing arrival times of the streamlines per material")
        reason_of_termination = reason_of_termination if reason_of_termination else config.streamline_reader.reason_of_termination
        temp_df = self.stream_data.copy()
        if reason_of_termination:
            temp_df = temp_df.filter(lambda x: x["ReasonForTermination"].max() == reason_of_termination)

        # temp_df['Material ID'] = temp_df['Material ID'].apply(np.ceil)
        temp_df['Material ID'] = (temp_df['Material ID'] + 0.45).apply(np.floor)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(1, 1.5), 1)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(1.5, 2.5), 2)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(2.5, 3.5), 3)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(3.5, 4.5), 4)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(4.5, 5), 5)
        temp_series: pd.Series = temp_df.groupby(["Material ID", "SeedIds"]).max()
        temp_series = temp_series.reset_index()

        # AUTOMATIC
        dic = {}
        dic_group = {}
        for group in temp_series.groupby('Material ID').groups:
            ngroup = temp_series.groupby('Material ID').get_group(group)['IntegrationTime']
            dic[group] = ngroup
            dic_group[group] = ''

        b = True
        for key in sorted(dic.keys(), reverse=True):
            if b:
                add = dic[key]
                dic_group[key] = add
                b = False
            else:
                add = pd.concat([add, dic[key]])
                dic_group[key] = add

        return temp_series, dic_group

    def compute_initial_velocities(self,
                                   reason_of_termination = None,
                                   normalize=True,
                                   index_df: pd.Series=None,
                                   ):
        """
        This method computes the initial velocities each streamlines 'sees' at the beginning, it can be used to normalize them later on

        Args:
            normalized: Parameter controlling weather the output vector should be normalized by dividing by the maximum velocity valuefdh

        Returns: a vector containing the initial velocities of the streamlines
        """
        logger.info("Computing initial velocities of the streamlines")
        temp_df = self.data
        if index_df is not None:
            seed_ids = index_df["SeedIds"]
            temp_df = temp_df[temp_df["SeedIds"].isin(seed_ids)]
        temp_df = temp_df.groupby("SeedIds").first()
        temp_series = temp_df["U:0"]
        if normalize:
            temp_series /= temp_series.sum()
        temp_series[temp_series < 0.0] = 0.0
        return temp_series

    def compute_length_streamlines(self, reason_of_termination=None) -> pd.Series:
        """
        This method computes the length of the streamlines
        Returns:
             A pd.Series object containing the length of the streamlines
        """
        logger.info("Computing length of the streamlines")
        reason_of_termination = reason_of_termination if reason_of_termination else config.streamline_reader.reason_of_termination
        temp_df = self.stream_data
        if reason_of_termination:
            temp_df = temp_df.filter(lambda x: x["ReasonForTermination"].max() == reason_of_termination)
        temp_series: pd.Series = temp_df.groupby("SeedIds").max()["arc_length"]
        return temp_series

    def compute_length_streamlines_per_material(self, reason_of_termination=None) -> pd.Series:
        """
        This method computes the length of the streamlines for a particular material
        Returns:
             A pd.Series object containing the length of the streamlines
        """
        logger.info("Computing length of the streamlines per material")
        reason_of_termination = reason_of_termination if reason_of_termination else config.streamline_reader.reason_of_termination
        temp_df = self.stream_data
        if reason_of_termination:
            temp_df = temp_df.filter(lambda x: x["ReasonForTermination"].max() == reason_of_termination)

        # temp_df['Material ID'] = temp_df['Material ID'].apply(np.ceil)
        temp_df['Material ID'] = (temp_df['Material ID'] + 0.45).apply(np.floor)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(1, 1.5), 1)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(1.5, 2.5), 2)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(2.5, 3.5), 3)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(3.5, 4.5), 4)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(4.5, 5), 5)

        temp_series: pd.Series = temp_df.groupby(["Material ID", "SeedIds"]).max()
        temp_series = temp_series.reset_index()

        # AUTOMATIC
        # dic = {}
        # for group in temp_series.groupby('Material ID').groups:
        #     ngroup_series = temp_series.groupby('Material ID').get_group(group)['arc_length']
            # temp_series = temp_series.groupby('Material ID').get_group(group)['arc_length']
            # dic[group] = ngroup

        return temp_series #, dic
        # return ngroup_series

    def compute_beta(self, aperture_field: str = None) -> pd.Series:
        """
        This method computes beta values for each streamline
        Returns:
            A pd.Series object containing the streamline info with the beta column added
        """
        self.number_of_zero_apertures = 0
        aperture_field_file = aperture_field if aperture_field else config.beta_integrator.aperture_field_file if config.beta_integrator.aperture_field_file else None
        assert aperture_field_file, "Define a file containing an aperture field matrix"
        # Read aperture field
        with open(aperture_field_file, "rb") as opened_file:
            aperture_field = pickle.load(opened_file)
        aperture_field = self.fix_aperture_field(aperture_field)
        logger.info(f"Aperture field has been loaded from {aperture_field_file}")
        logger.info("Computing beta values for the streamlines")
        self.data["beta"] = 0.0
        # Filter streamlines
        temp_df = self.filter_streamlines()

        for stream in tqdm(temp_df.groups):
            # if random.random() > 0.01:
            #     continue
            stream_data = self.stream_data.get_group(stream)
            index_group = stream_data.index
            stream_beta_value = self.integrate_beta(stream=stream_data, aperture_field=aperture_field)
            if stream_beta_value == 0.0:
                continue
            if self.is_aperture_zero:
                self.is_aperture_zero = False
                continue
            else:
                self.data.loc[index_group, "beta"] = stream_beta_value

        return self.data.groupby('SeedIds').max()["beta"]

    def integrate_beta(self, stream: pd.DataFrame, aperture_field: np.ndarray):
        """
        This method integrates a given streamline to compute the value of beta
        Args:
            stream: A pandas DataFrame containing the data of a given streamline
            aperture_field: A matrix containing the information of the aperture field
        Returns:

        """
        beta = 0.0
        aperture_field_nx = aperture_field.shape[1]
        aperture_field_ny = aperture_field.shape[0]
        previous_integration_time = 0.0
        previous_aperture = 0.0
        for index, fragment in stream.iterrows():
            # aperture = aperture_from_a_xy_point(x_point=)
            # Nearest neighbour
            index_x = int(np.floor(fragment["x"] / config.beta_integrator.dimension_x * aperture_field_nx))
            index_y = int(np.floor(fragment["y"] / config.beta_integrator.dimension_y * aperture_field_ny))
            index_row = index_y
            index_column = index_x
            if index_row == aperture_field_ny:
                index_row -= 1
            if index_column == aperture_field_nx:
                index_column -= 1

            # Compute tau
            tau = fragment["IntegrationTime"] - previous_integration_time
            previous_integration_time = fragment["IntegrationTime"]
            # if tau < 0.0:
            #     print(tau)

            # Calculate aperture
            aperture = aperture_field[index_row, index_column]
            # if fragment["x"] < 0.001:
            #     print(f"Aperture: {aperture}, x: {fragment['x']} y: {fragment['y']} index_x: {index_x}, index_y: {index_y}")
            if aperture == 0.0:
                aperture = previous_aperture
                continue
            previous_aperture = aperture
            # print(f"Tau: {tau / (365 * 24)} h, Aperture: {aperture} m, Fragment beta: {2 * tau / aperture / (365 * 24 * 3600)}, Cummulated beta: {beta}")
            beta += 2 * tau / aperture / (365 * 24 * 3600)
        # print(f"Stream Finishes")
        # print(f"Computed beta: {beta}")
        return beta

    def get_data(self) -> np.ndarray:
        """
        Outputs the data
        :return: np.ndarray object containing centroid information and variable output
        """
        return self.data.values

    def dump_to_csv(self, output_file, delimiter=","):
        """
        Writes the data into a csv file
        :param output_file:
        :return:
        """
        print(f"Starting dump into {output_file}")
        # self.data.to_csv(output_file, delimiter=delimiter)
        self.data.to_csv(output_file)
        print(f"The data has been properly exported to the {output_file} file")

    def to_csv(self, output_file, delimiter=","):
        """
        Writes the data into a csv file
        :param output_file:
        :return:
        """
        print(f"Starting dump into {output_file}")
        # self.data.to_csv(output_file, delimiter=delimiter)
        self.data.to_csv(output_file)
        print(f"The data has been properly exported to the {output_file} file")

    def filter_streamlines(self,
                           reason_of_termination=None,
                           min_x=None,
                           min_y=None,
                           ) -> DataFrameGroupBy:


        reason_of_termination = reason_of_termination if reason_of_termination else config.streamline_reader.filter.reason_of_termination if config.streamline_reader.filter.reason_of_termination else None
        min_x = min_x if min_x else config.streamline_reader.filter.min_x if config.streamline_reader.filter.min_x else None
        logger.info("Filtering streamlines")
        temp_df = self.data.groupby("SeedIds")
        initial_number_of_streamlines = len(temp_df.groups)
        if reason_of_termination:
            temp_df = temp_df.filter(lambda x: x.max()["ReasonForTermination"] == float(reason_of_termination))
            temp_df = temp_df.groupby("SeedIds")
            logger.info(f"{initial_number_of_streamlines - len(temp_df.groups)} streamlines have been filtered due to 'Reason of termination' filtering (Keeping {len(temp_df.groups) / initial_number_of_streamlines * 100: 2.1f}% of total)")
            # initial_number_of_streamlines = temp_df.shape[0]
        if min_x is not None:
            temp_df = temp_df.filter(lambda x: x.max()["x"] > min_x)
            temp_df = temp_df.groupby("SeedIds")
            logger.info(f"{initial_number_of_streamlines - len(temp_df.groups)} streamlines have been filtered due to 'min_x' filtering (Keeping {len(temp_df.groups) / initial_number_of_streamlines * 100:2.1f}% of total)")

        if min_y is not None:
            temp_df = temp_df.filter(lambda x: x.max()["y"] > min_y)
            temp_df = temp_df.groupby("SeedIds")
            logger.info(f"{initial_number_of_streamlines - len(temp_df.groups)} streamlines have been filtered due to 'min_y' filtering (Keeping {len(temp_df.groups) / initial_number_of_streamlines * 100:2.1f}% of total)")
        return temp_df

    @staticmethod
    def fix_aperture_field(aperture_matrix):
        aperture_matrix[0, :] = aperture_matrix[1, :]
        aperture_matrix[:, 0] = aperture_matrix[:, 1]
        aperture_matrix[aperture_matrix.shape[0] - 1, :] = aperture_matrix[aperture_matrix.shape[0] - 2, :]
        aperture_matrix[:, aperture_matrix.shape[1] - 1] = aperture_matrix[:, aperture_matrix.shape[1] - 2]
        return aperture_matrix

    def integrate_variable_within_materials(self, variable,
                                            material_names=None,
                                            add_variable_name_to_output=True,
                                            output_variable_name = None,
                                            *args,
                                            **kwargs
                                            ):
        logger.info(f'Integrating variable {variable} within the materials for all the streamlines')
        output_list = []
        for streamline in self.stream_data:
            output_list.append(self._integrate_variable_within_materials_single(streamline=streamline[1],
                                                           variable=variable,
                                                           *args,
                                                           **kwargs)
                             )
        output_df = pd.DataFrame(output_list)
        if material_names:
            if add_variable_name_to_output:
                if not output_variable_name:
                    output_variable_name = variable
                output_column_names = [f"{material_names[column]}-{output_variable_name}" for column in output_df.columns]
            else:
                output_column_names = [material_names[column] for column in output_df.columns]
            output_df.columns = output_column_names
        return output_df


    @staticmethod
    def _integrate_variable_within_materials_single(streamline: pd.DataFrame,
                                            variable: str,
                                            categorize_materials: bool=True,
                                            material_id_column: str='Material ID',
                                            material_id_offset: float = 0.45,
                                            ) -> dict:
        """
        This method implements an algorithm to integrate a given variable along the different materials of a streamline
        Args:
            variable: variable to integrate

        Returns:
            pd.Dataframe containing the materials and the integrated variable
        """
        streamline = streamline.reset_index()
        if categorize_materials:
            streamline[material_id_column] = (streamline[material_id_column] + material_id_offset).apply(np.floor)
        material_groups = streamline.groupby(material_id_column).groups
        output_dict = {material: 0.0 for material in material_groups.keys()}
        var_current = None
        for idx, segment in streamline.iterrows():
            if idx == 0:
                # Initial segment
                output_dict[segment[material_id_column]] += segment[variable]
                var_current = segment[variable]
                continue
            if var_current == None:
                raise ValueError('There was an error when processing the initial segment')
            output_dict[segment[material_id_column]] += segment[variable] - var_current
            var_current = segment[variable]
        return output_dict


import os
import numpy as np
from ..preprocessing.mesh_preprocessor import MeshPreprocessor
import logging
import math
logger = logging.getLogger(__name__)
import meshio as msh
from tqdm import tqdm


class FemReader(MeshPreprocessor):
    upscaled_permeability = {} # dict of upscaled permeability values
    upscaled_porosity = {} # dict of upscaled porosity values

    def __init__(self, filename, kd_tree=True, st_file=False):
        super().__init__(st_file=False)
        self.aux_nodes = []
        self.open_file(filename)
        # Create KD-mesh structure
        if kd_tree:
            self.create_kd_tree()

    def open_file(self, filename):
        if self.is_streamlit:
            import streamlit as st
        with open(filename, "r") as f:
            for line in f:
                line = line.rstrip()
                split_line = line.split()


                if split_line[0] == "CLASS":
                    line = f.readline()
                    line = line.rstrip()
                    split_line = line.split()
                    self.aux_n_layers = int(split_line[4])

                if split_line[0] == "DIMENS":
                    line = f.readline()
                    line = line.rstrip()
                    split_line = line.split()
                    self.aux_n_nodes = int(split_line[0])
                    self.aux_n_elements = int(split_line[1])
                    self.aux_num_nodes_per_elem = int(split_line[2])

                elif split_line[0] == "NODE":
                    nodes_elem = np.zeros([self.aux_n_elements, self.aux_num_nodes_per_elem], dtype=int)
                    for e in tqdm(range(0, self.aux_n_elements), desc='Reading elements'):
                        line = f.readline()
                        line = line.rstrip()
                        split_line = line.split()
                        for ne in range(0, self.aux_num_nodes_per_elem):
                            nodes_elem[e, ne] = int(split_line[ne]) - 1

                elif split_line[0] == "VARNODE":
                    line = f.readline()
                    line = line.rstrip()
                    split_line = line.split()
                    num_elem_per_node = int(split_line[1])
                    nodes_elem = np.zeros([self.aux_n_elements, num_elem_per_node], dtype=int)
                    element_type = np.zeros([self.aux_n_elements], dtype=int)

                    for e in tqdm(range(0, self.aux_n_elements), ):
                        line = f.readline()
                        line = line.rstrip()
                        split_line = line.split()
                        for ne in range(0, num_elem_per_node):
                            nodes_elem[e, ne] = int(split_line[ne + 1]) - 1
                            element_type[e] = int(split_line[0])


                elif split_line[0] == "COOR":
                    line = f.readline()
                    line = line.rstrip().replace(',', '').split()
                    # for np in tqdm(range(0, self.aux_n_nodes), desc='Reading nodes'):

                    self.aux_xcoor = []
                    self.aux_ycoor = []
                    #self.aux_zcoor = np.zeros([self.aux_n_nodes])  #NOR DEFINITIVE, WE NEED TO READ ELEVATIONS BEFORE CRETING THE NODES.
                    for l in range(0, self.aux_n_nodes/(self.aux_n_layers+1)):
                        for i in range(0,12):
                            self.aux_xcoor.append(float(line[i]))
                    for l in range(0, self.aux_n_nodes/(self.aux_n_layers+1)):
                        for i in range(0,12):
                            self.aux_ycoor.append(float(line[i]))
                    for n in range(0, self.aux_n_nodes):
                        self.aux_nodes.append(np.array([self.aux_xcoor[n], self.aux_ycoor[n], self.aux_zcoor[n]]))
                        #line = f.readline()
                        #line = line.rstrip().replace(',', '').split()
                    break

                elif split_line[0] == "XYZCOOR":
                    line = f.readline()
                    line = line.rstrip().replace(',', '').split()

                    for n in tqdm(range(0, self.aux_n_nodes), desc='Reading nodes'):
                        self.aux_nodes.append(np.array([float(line[0]), float(line[1]), float(line[2])]))
                        line = f.readline()
                        line = line.rstrip().replace(',', '').split()
                    break



                elif split_line[0] == "ELEV_I":
                    line = f.readline()
                    line = line.rstrip().replace(',', '').split()

                    self.aux_elevation_value = []
                    self.nodes_with_this_elevation = []

                    #For each layer:
                        #LOOP TO READ ELEVATION VALUES [Line[0]]
                        #READ ELEVATION VALUE, AND MATRIX WITH MAXIMUM OF 13 COLUMNS HAVING ALL NODES WITH THAT ELEVATION.
                        #                          [Line[1] to Line[14] for I don't know how many lines]

                    break

                else:
                    continue

        if self.is_streamlit:
            st.write(f'Creating {self.aux_n_elements} mesh elements')
            progress_tetra = st.empty()
            progress_tetra.progress(0)
            count = 0


        for e in range(0, self.aux_n_elements):
            if self.is_streamlit:
                proportion = round(count / self.aux_n_elements * 100)
                progress_tetra.progress(proportion)
                count += 1

            if element_type[2] == 6:
                # Add a tetrahedra to the mesh structure
                self.add_tetrahedra(node_ids=nodes_elem[e],
                                    node_coords=[
                                        self.aux_nodes[nodes_elem[e, 0]],
                                        self.aux_nodes[nodes_elem[e, 1]],
                                        self.aux_nodes[nodes_elem[e, 2]],
                                        self.aux_nodes[nodes_elem[e, 3]],
                                    ]
                                    )

            else:
                logger.warning(f"Element type {element_type[e]} not supported")










"""
Global variables
"""
import yaml
from box import Box
from pathlib import Path


def get_root_path() -> Path:
    """Returns path to the root of the project"""
    return Path(__file__).parent.parent


def get_config_path() -> Path:
    """Returns path to the root of the project"""
    return Path(__file__).parent.parent / "config"


def test_data_path() -> Path:
    """Returns path to the root of the project"""
    return Path(__file__).parent.parent / "tests/test_data"


def runtime_path():
    return Path.cwd()


def read_config(config_file="./config.yaml"):
    """
    Reads the configuration file
    :param config_file:
    :return:
    """
    with open(config_file) as file:
        context = yaml.load(file, Loader=yaml.FullLoader)
    return Box(context, default_box=True)
    # return munch.DefaultMunch.from_dict(context)


def initialize_config(config_file="./config/config.yaml"):
    global config
    try:
        config = read_config(config_file)
    except FileNotFoundError as fe:
        print(f"ERROR: Config file({config_file}) not found.")
        exit(1)




import numpy as np

from pydelling.config import config
from .BaseFace import BaseFace
import pandas as pd
from .BaseAbstractMeshObject import BaseAbstractMeshObject
from typing import *
from natsort import natsorted
import sympy as sp
from pydelling.utils.geometry import Point, Plane, Line
from itertools import product, combinations
from pydelling.preprocessing.dfn_preprocessor.Fracture import Fracture
from pydelling.utils.geometry_utils import filter_unique_points


class BaseElement(BaseAbstractMeshObject):
    local_id = 0
    eps = 1E-13
    __slots__ = ['node_ids', 'node_coords']

    def __init__(self, node_ids, node_coords, centroid_coords=None):
        self.nodes: np.ndarray = node_ids  # Node id set
        self.coords: np.ndarray = np.array(node_coords)  # Coordinates of each node
        self.centroid_coords = centroid_coords
        self.local_id = BaseElement.local_id # Element id
        BaseElement.local_id += 1
        self.faces: Dict[str, BaseFace] = {}  # Faces of an element
        self.type = None
        self.meshio_type = None
        self.associated_fractures = {}
        self.associated_faults = {}
        self.total_fracture_volume = 0
        self.area = 0

    def __repr__(self):
        return f"{self.type} {self.local_id}"

    def print_element_info(self):
        print("### Element info ###")
        print(f"Element ID: {self.local_id}")
        print(f"Number of nodes: {self.n_nodes}")
        print(f"Element type: {self.type}")
        print(f"Node list: {self.nodes}")
        print("### End element info ###")

    def print_face_info(self):
        print("### Face info ###")
        for face in self.faces:
            print(f"{face}: {self.faces[face].nodes}")
        print("### End face info ###")

    def intersect_faces_with_plane(self, plane: Plane):
        """Returns the intersection of a face with a plane

        Args:
            face: Face to intersect
            plane: Plane to intersect with

        Returns: Intersection points
        """
        intersected_lines = []
        intersected_points = []

        for face in self.faces:
            intersection = self.faces[face].intersect_with_plane(plane)
            if intersection:
                intersected_lines.append(intersection)


        # Intersect the lines within themselves
        intersected_points = list(self._full_line_intersections(intersected_lines=intersected_lines))
        # Check what happens with the fracture points

        # Check if the intersected points are inside the element
        intersected_inside_points = []
        for point in intersected_points:
            if self.contains(point):
                intersected_inside_points.append(point)
        intersected_points = intersected_inside_points.copy()

        return intersected_points


    def intersect_with_fracture(self, fracture: Fracture):
        """Intersects an element with a fracture"""
        intersected_lines = []
        intersected_points = []

        for face in self.faces:
            intersection = self.faces[face].intersect_with_plane(fracture.plane)
            for corner_line in fracture.corner_lines:
                line_intersection = corner_line.intersect(self.faces[face].plane)
                if line_intersection is not None:
                    # if self.contains(line_intersection):
                    intersected_points.append(line_intersection)

            if intersection:
                intersected_lines.append(intersection)

        # Intersect the lines within themselves
        intersected_points += list(self._full_line_intersections(intersected_lines=intersected_lines))

        # Check if the intersected points are inside the element
        intersected_inside_points = []
        for point in intersected_points:
            if self.contains(point):
                intersected_inside_points.append(point)
        intersected_points = intersected_inside_points.copy()

        # Test algorithm that checks if a point is contained in the fracture
        final_points = []
        for point in intersected_points:
            if fracture.contains(point):
                final_points.append(point)

        # Check if any fracture edge is inside the element. If so, add that as intersection point
        for corner in fracture.corners:
            if self.contains(corner):
                final_points.append(corner)

        # final_points = intersected_points
        final_points = filter_unique_points(final_points)
        final_points = [Point(point) for point in final_points]

        return final_points

    def _full_line_intersections(self, intersected_lines: List[Line]) -> List:
        """Intersects a list of lines with each other"""
        intersected_points = []
        line_combination = list(combinations(intersected_lines, 2))
        for line_pair in line_combination:
            line_1: Line = line_pair[0]
            line_2: Line = line_pair[1]
            intersection = line_1.intersect(line_2)
            if intersection is not None:
                intersected_points.append(intersection)
        return intersected_points

    def contains(self, point: np.ndarray or Point) -> bool:
        """Checks if a point is inside the element"""
        for face in self.faces:
            dot_plane_point = np.dot(self.faces[face].unit_normal_vector, point - self.faces[face].centroid)
            if dot_plane_point > self.eps:
                return False
        return True


    @property
    def n_nodes(self):
        return len(self.nodes)

    @property
    def edges(self):
        edges_list = []
        for face in self.faces:
            face_edges = self.faces[face].edges
            edges_list.append(face_edges)

        # Delete duplicates
        edge_list_flatten = [sorted(item) for sublist in edges_list for item in sublist]
        edge_list_unique = np.unique(edge_list_flatten, axis=0)
        return edge_list_unique

    @property
    def volume(self):
        """Returns the volume of the hexahedra

        Returns: volume of the hexahedra
        """
        return None




from scipy.spatial.qhull import ConvexHull

from pydelling.readers.iGPReader.geometry import QuadrilateralFace, TriangleFace, BaseElement
from pydelling.readers.iGPReader.utils.geometry_utils import *


class WedgeElement(BaseElement):
    def __init__(self, node_ids, node_coords, element_type_n, local_id, centroid_coords=None):
        super().__init__(node_ids, node_coords, element_type_n, local_id, centroid_coords)
        self.define_faces()  # Define faces of the element
        # self.centroid = self.compute_centroid()
        # self.centroid_coords = self.centroid

        if centroid_coords is None:
            self.centroid = self.compute_centroid()
            self.centroid_coords = self.centroid
        else:
            self.centroid = np.array(centroid_coords)
            self.centroid_coords = self.centroid

    def define_faces(self):
        # Add faces that define the wedge
        # Face 1
        self.faces["q1"] = QuadrilateralFace(nodes=np.array([self.nodes[0],
                                                             self.nodes[1],
                                                             self.nodes[4],
                                                             self.nodes[3]]),
                                             coords=np.array([self.coords[0],
                                                              self.coords[1],
                                                              self.coords[4],
                                                              self.coords[3]]))

        # Face 2
        self.faces["q2"] = QuadrilateralFace(nodes=np.array([self.nodes[1],
                                                             self.nodes[2],
                                                             self.nodes[5],
                                                             self.nodes[4]]),
                                             coords=np.array([self.coords[1],
                                                              self.coords[2],
                                                              self.coords[5],
                                                              self.coords[4]]))
        # Face 3
        self.faces["q3"] = QuadrilateralFace(nodes=np.array([self.nodes[2],
                                                             self.nodes[0],
                                                             self.nodes[3],
                                                             self.nodes[5]]),
                                             coords=np.array([self.coords[2],
                                                              self.coords[0],
                                                              self.coords[3],
                                                              self.coords[5]]))
        # Face 4
        self.faces["t1"] = TriangleFace(nodes=np.array([self.nodes[0],
                                                        self.nodes[2],
                                                        self.nodes[1]]),
                                        coords=np.array([self.coords[0],
                                                         self.coords[2],
                                                         self.coords[1]]))
        # Face 5
        self.faces["t2"] = TriangleFace(nodes=np.array([self.nodes[3],
                                                        self.nodes[4],
                                                        self.nodes[5]]),
                                        coords=np.array([self.coords[3],
                                                         self.coords[4],
                                                         self.coords[5]]))

    def compute_volume(self):
        """
        Computes volume of a general polyhedra based on the convex hull of a set of points
        :return: volume of the polyhedron
        """
        return ConvexHull(self.coords).volume

    def compute_centroid(self):
        """
        Computes the centroid of a general polyhedra
        :return: centroid of the polyhedron
        """
        return np.mean(self.coords, axis=0)

    @property
    def volume(self):
        """Returns the volume of the hexahedra

        Returns: volume of the hexahedra
        """
        return self.compute_volume()